@@ -17,7 +17,7 @@ from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
 from trytond import backend
 from trytond.pyson import Eval, Bool, If, PYSONEncoder, Id
 from trytond.transaction import Transaction
-from trytond.pool import Pool, PoolMeta
+from trytond.pool import Pool
 
 from trytond.modules.account.tax import TaxableMixin
 from trytond.modules.product import price_digits
@@ -27,7 +27,6 @@ __all__ = ['Purchase', 'PurchaseIgnoredInvoice',
     'PurchaseLineIgnoredMove', 'PurchaseLineRecreatedMove', 'PurchaseReport',
     'OpenSupplier', 'HandleShipmentExceptionAsk', 'HandleShipmentException',
     'HandleInvoiceExceptionAsk', 'HandleInvoiceException']
-__metaclass__ = PoolMeta
 
 _STATES = {
     'readonly': Eval('state') != 'draft',
@@ -39,7 +38,7 @@ _ZERO = Decimal(0)
 class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     'Purchase'
     __name__ = 'purchase.purchase'
-    _rec_name = 'reference'
+    _rec_name = 'number'
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
@@ -49,8 +48,8 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                 Eval('context', {}).get('company', -1)),
             ],
         depends=['state'], select=True)
-    reference = fields.Char('Reference', size=None, readonly=True, select=True)
-    supplier_reference = fields.Char('Supplier Reference', select=True)
+    number = fields.Char('Number', size=None, readonly=True, select=True)
+    reference = fields.Char('Reference', select=True)
     description = fields.Char('Description', size=None, states=_STATES,
         depends=_DEPENDS)
     state = fields.Selection([
@@ -227,7 +226,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         Move = pool.get('stock.move')
         InvoiceLine = pool.get('account.invoice.line')
         TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         model_data = Table('ir_model_data')
         model_field = Table('ir_model_field')
         sql_table = cls.__table__()
@@ -254,9 +253,14 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                         len('packing'))],
                 where=model_field.name.like('%packing%')
                 & (model_field.module == module_name)))
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.column_rename('packing_state', 'shipment_state')
 
+        # Migration from 3.8:
+        if table.column_exist('supplier_reference'):
+            table.column_rename('reference', 'number')
+            table.column_rename('supplier_reference', 'reference')
+
         super(Purchase, cls).__register__(module_name)
 
         # Migration from 1.2: rename packing to shipment in
@@ -266,7 +270,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                 values=['shipment'],
                 where=sql_table.invoice_method == 'packing'))
 
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         # Migration from 2.2: warehouse is no more required
         table.not_null_action('warehouse', 'remove')
 
@@ -275,9 +279,9 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
 
         # Migration from 3.2
         # state confirmed splitted into confirmed and processing
-        if (TableHandler.table_exist(cursor, PurchaseLine._table)
-                and TableHandler.table_exist(cursor, Move._table)
-                and TableHandler.table_exist(cursor, InvoiceLine._table)):
+        if (TableHandler.table_exist(PurchaseLine._table)
+                and TableHandler.table_exist(Move._table)
+                and TableHandler.table_exist(InvoiceLine._table)):
             purchase_line = PurchaseLine.__table__()
             move = Move.__table__()
             invoice_line = InvoiceLine.__table__()
@@ -302,7 +306,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                     where=sql_table.id.in_(sub_query.select(sub_query.id))))
 
         # Add index on create_date
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.index_action('create_date', action='add')
 
     @classmethod
@@ -362,7 +366,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     def on_change_party(self):
         pool = Pool()
         Currency = pool.get('currency.currency')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         table = self.__table__()
         self.invoice_address = None
         self.payment_term = self.default_payment_term()
@@ -445,10 +449,6 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                     if value is not None else default_value,)
         return taxable_lines
 
-    @property
-    def tax_type(self):
-        return 'invoice'
-
     def get_tax_amount(self):
         taxes = self._get_taxes().itervalues()
         return sum(tax['amount'] for tax in taxes)
@@ -572,20 +572,20 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                     })
 
     def get_rec_name(self, name):
-        return (self.reference or str(self.id)
+        return (self.number or str(self.id)
             + ' - ' + self.party.name)
 
     @classmethod
     def search_rec_name(cls, name, clause):
         _, operator, value = clause
-        if operator[1].startswith('!') or operator[1].startswith('not '):
+        if operator.startswith('!') or operator.startswith('not '):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
         names = value.split(' - ', 1)
         domain = [bool_op,
+            ('number', operator, names[0]),
             ('reference', operator, names[0]),
-            ('supplier_reference', operator, names[0]),
             ]
         if len(names) != 1 and names[1]:
             domain = [bool_op, domain, ('party', operator, names[1])]
@@ -601,7 +601,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             default = {}
         default = default.copy()
         default['state'] = 'draft'
-        default['reference'] = None
+        default['number'] = None
         default['invoice_state'] = 'none'
         default['invoices_ignored'] = None
         default['moves'] = None
@@ -617,9 +617,9 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                 self.raise_user_error('warehouse_required', (self.rec_name,))
 
     @classmethod
-    def set_reference(cls, purchases):
+    def set_number(cls, purchases):
         '''
-        Fill the reference field with the purchase sequence
+        Fill the number field with the purchase sequence
         '''
         pool = Pool()
         Sequence = pool.get('ir.sequence')
@@ -627,12 +627,10 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
 
         config = Config(1)
         for purchase in purchases:
-            if purchase.reference:
+            if purchase.number:
                 continue
-            reference = Sequence.get_id(config.purchase_sequence.id)
-            cls.write([purchase], {
-                    'reference': reference,
-                    })
+            purchase.number = Sequence.get_id(config.purchase_sequence.id)
+        cls.save(purchases)
 
     @classmethod
     def set_purchase_date(cls, purchases):
@@ -652,21 +650,8 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                     'total_amount_cache': purchase.total_amount,
                     })
 
-    def _get_invoice_line_purchase_line(self, invoice_type):
-        '''
-        Return invoice line for each purchase lines
-        '''
-        res = {}
-        for line in self.lines:
-            val = line.get_invoice_line(invoice_type)
-            if val:
-                res[line.id] = val
-        return res
-
-    def _get_invoice_purchase(self, invoice_type):
-        '''
-        Return invoice of type invoice_type
-        '''
+    def _get_invoice_purchase(self):
+        'Return invoice'
         pool = Pool()
         Journal = pool.get('account.journal')
         Invoice = pool.get('account.invoice')
@@ -681,7 +666,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
 
         return Invoice(
             company=self.company,
-            type=invoice_type,
+            type='in',
             journal=journal,
             party=self.party,
             invoice_address=self.invoice_address,
@@ -690,10 +675,8 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             payment_term=self.payment_term,
             )
 
-    def create_invoice(self, invoice_type):
-        '''
-        Create an invoice for the purchase and return it
-        '''
+    def create_invoice(self):
+        'Create an invoice for the purchase and return it'
         pool = Pool()
         Invoice = pool.get('account.invoice')
 
@@ -704,11 +687,15 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             self.raise_user_error('missing_account_payable',
                     error_args=(self.party.rec_name,))
 
-        invoice_lines = self._get_invoice_line_purchase_line(invoice_type)
+        invoice_lines = {}
+        for line in self.lines:
+            ilines = line.get_invoice_line()
+            if ilines:
+                invoice_lines[line.id] = ilines
         if not invoice_lines:
             return
 
-        invoice = self._get_invoice_purchase(invoice_type)
+        invoice = self._get_invoice_purchase()
         invoice.lines = list(chain.from_iterable(
                 invoice_lines[l.id] for l in self.lines
                 if l.id in invoice_lines))
@@ -721,21 +708,31 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         '''
         Create move for each purchase lines
         '''
-        new_moves = []
+        pool = Pool()
+        Move = pool.get('stock.move')
+
+        moves = []
         for line in self.lines:
-            if (line.quantity >= 0) != (move_type == 'in'):
-                continue
-            move = line.create_move()
+            move = line.get_move(move_type)
             if move:
-                new_moves.append(move)
-        return new_moves
+                moves.append(move)
+        Move.save(moves)
+        return moves
+
+    @property
+    def return_from_location(self):
+        if self.warehouse:
+            return (self.warehouse.supplier_return_location
+                or self.warehouse.storage_location)
 
     def _get_return_shipment(self):
         ShipmentInReturn = Pool().get('stock.shipment.in.return')
         return ShipmentInReturn(
             company=self.company,
-            from_location=self.warehouse.storage_location,
+            from_location=self.return_from_location,
             to_location=self.party.supplier_location,
+            supplier=self.party,
+            delivery_address=self.party.address_get(type='delivery'),
             )
 
     def create_return_shipment(self, return_moves):
@@ -784,7 +781,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     def quote(cls, purchases):
         for purchase in purchases:
             purchase.check_for_quotation()
-        cls.set_reference(purchases)
+        cls.set_number(purchases)
 
     @classmethod
     @ModelView.button
@@ -798,6 +795,11 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     def proceed(cls, purchases):
         pass
 
+    @classmethod
+    @Workflow.transition('done')
+    def do(cls, purchases):
+        pass
+
     @classmethod
     @ModelView.button_action('purchase.wizard_invoice_handle_exception')
     def handle_invoice_exception(cls, purchases):
@@ -813,8 +815,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     def process(cls, purchases):
         process, done = [], []
         for purchase in purchases:
-            purchase.create_invoice('in_invoice')
-            purchase.create_invoice('in_credit_note')
+            purchase.create_invoice()
             purchase.set_invoice_state()
             purchase.create_move('in')
             return_moves = purchase.create_move('return')
@@ -828,9 +829,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         if process:
             cls.proceed(process)
         if done:
-            cls.write(done, {
-                    'state': 'done',
-                    })
+            cls.do(done)
 
 
 class PurchaseIgnoredInvoice(ModelSQL):
@@ -875,6 +874,7 @@ class PurchaseLine(ModelSQL, ModelView):
             },
         depends=['unit_digits', 'type'])
     unit = fields.Many2One('product.uom', 'Unit',
+        ondelete='RESTRICT',
         states={
             'required': Bool(Eval('product')),
             'invisible': Eval('type') != 'line',
@@ -889,7 +889,7 @@ class PurchaseLine(ModelSQL, ModelView):
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
         'on_change_with_unit_digits')
     product = fields.Many2One('product.product', 'Product',
-        domain=[('purchasable', '=', True)],
+        ondelete='RESTRICT', domain=[('purchasable', '=', True)],
         states={
             'invisible': Eval('type') != 'line',
             'readonly': ~Eval('_parent_purchase'),
@@ -970,10 +970,10 @@ class PurchaseLine(ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         sql_table = cls.__table__()
         super(PurchaseLine, cls).__register__(module_name)
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
 
         # Migration from 1.0 comment change into note
         if table.column_exist('comment'):
@@ -997,7 +997,7 @@ class PurchaseLine(ModelSQL, ModelView):
     def get_move_done(self, name):
         Uom = Pool().get('product.uom')
         done = True
-        if not self.product:
+        if self.type != 'line' or not self.product:
             return True
         if self.product.type == 'service':
             return True
@@ -1162,13 +1162,13 @@ class PurchaseLine(ModelSQL, ModelView):
         return Decimal('0.0')
 
     def get_from_location(self, name):
-        if self.quantity >= 0:
+        if (self.quantity or 0) >= 0:
             return self.purchase.party.supplier_location.id
-        elif self.purchase.warehouse:
-            return self.purchase.warehouse.storage_location.id
+        elif self.purchase.return_from_location:
+            return self.purchase.return_from_location.id
 
     def get_to_location(self, name):
-        if self.quantity >= 0:
+        if (self.quantity or 0) >= 0:
             if self.purchase.warehouse:
                 return self.purchase.warehouse.input_location.id
         else:
@@ -1178,9 +1178,14 @@ class PurchaseLine(ModelSQL, ModelView):
         '_parent_purchase.purchase_date', '_parent_purchase.party')
     def on_change_with_delivery_date(self, name=None):
         if self.moves:
-            return min(m.effective_date if m.effective_date else m.planned_date
-                for m in self.moves)
+            dates = filter(
+                None, (m.effective_date or m.planned_date for m in self.moves))
+            if dates:
+                return min(dates)
+            else:
+                return
         if (self.product
+                and self.quantity is not None
                 and self.quantity > 0
                 and self.purchase and self.purchase.party
                 and self.product.product_suppliers):
@@ -1193,12 +1198,9 @@ class PurchaseLine(ModelSQL, ModelView):
                         return None
                     return delivery_date
 
-    def get_invoice_line(self, invoice_type):
-        '''
-        Return a list of invoice line for purchase line
-        '''
+    def get_invoice_line(self):
+        'Return a list of invoice line for purchase line'
         pool = Pool()
-        Uom = pool.get('product.uom')
         Property = pool.get('ir.property')
         InvoiceLine = pool.get('account.invoice.line')
 
@@ -1208,62 +1210,26 @@ class PurchaseLine(ModelSQL, ModelView):
         invoice_line.note = self.note
         invoice_line.origin = self
         if self.type != 'line':
-            if (self.purchase.invoice_method == 'order'
-                    and not self.invoice_lines
-                    and ((all(l.quantity >= 0 for l in self.purchase.lines
-                                if l.type == 'line')
-                            and invoice_type == 'in_invoice')
-                        or (all(l.quantity <= 0 for l in self.purchase.lines
-                                if l.type == 'line')
-                            and invoice_type == 'in_credit_note'))):
+            if self._get_invoice_not_line():
                 return [invoice_line]
             else:
                 return []
-        if (invoice_type == 'in_invoice') != (self.quantity >= 0):
-            return []
 
-        stock_moves = []
-        if (self.purchase.invoice_method == 'order'
-                or not self.product
-                or self.product.type == 'service'):
-            quantity = abs(self.quantity)
-            stock_moves = self.moves
-        else:
-            quantity = 0.0
-            for move in self.moves:
-                if move.state == 'done':
-                    quantity += Uom.compute_qty(move.uom, move.quantity,
-                        self.unit)
-                    if move.invoiced_quantity < move.quantity:
-                        stock_moves.append(move)
-        invoice_line.stock_moves = stock_moves
+        quantity = (self._get_invoice_line_quantity()
+            - self._get_invoiced_quantity())
 
-        skip_ids = set(l.id for i in self.purchase.invoices_recreated
-            for l in i.lines)
-        for old_invoice_line in self.invoice_lines:
-            if old_invoice_line.type != 'line':
-                continue
-            if old_invoice_line.id not in skip_ids:
-                if old_invoice_line.invoice:
-                    old_invoice_type = old_invoice_line.invoice.type
-                else:
-                    old_invoice_type = old_invoice_line.invoice_type
+        if self.unit:
+            quantity = self.unit.round(quantity)
+        invoice_line.quantity = quantity
 
-                sign = 1 if old_invoice_type == invoice_type else -1
-
-                quantity -= Uom.compute_qty(old_invoice_line.unit,
-                        sign * old_invoice_line.quantity, self.unit)
-
-        rounding = self.unit.rounding if self.unit else 0.01
-        invoice_line.quantity = Uom.round(quantity, rounding)
-        if invoice_line.quantity <= 0:
+        if not invoice_line.quantity:
             return []
 
         invoice_line.unit = self.unit
         invoice_line.product = self.product
         invoice_line.unit_price = self.unit_price
         invoice_line.taxes = self.taxes
-        invoice_line.invoice_type = invoice_type
+        invoice_line.invoice_type = 'in'
         if self.product:
             invoice_line.account = self.product.account_expense_used
             if not invoice_line.account:
@@ -1279,31 +1245,65 @@ class PurchaseLine(ModelSQL, ModelView):
             if not invoice_line.account:
                 self.raise_user_error('missing_account_expense_property',
                     {'purchase': self.purchase.rec_name})
+        invoice_line.stock_moves = self._get_invoice_line_moves()
         return [invoice_line]
 
-    @classmethod
-    def view_attributes(cls):
-        return [
-            ('/form//field[@name="note"]|/form//field[@name="description"]',
-                'spell', Eval('_parent_purchase', {}).get('party_lang'))]
+    def _get_invoice_not_line(self):
+        'Return if the not line should be invoiced'
+        return (self.purchase.invoice_method == 'order'
+            and not self.invoice_lines)
 
-    @classmethod
-    def copy(cls, lines, default=None):
-        if default is None:
-            default = {}
-        default = default.copy()
-        default['moves'] = None
-        default['moves_ignored'] = None
-        default['moves_recreated'] = None
-        default['invoice_lines'] = None
-        return super(PurchaseLine, cls).copy(lines, default=default)
+    def _get_invoice_line_quantity(self):
+        'Return the quantity that should be invoiced'
+        pool = Pool()
+        Uom = pool.get('product.uom')
 
-    def get_move(self):
+        if (self.purchase.invoice_method == 'order'
+                or not self.product
+                or self.product.type == 'service'):
+            return self.quantity
+        elif self.purchase.invoice_method == 'shipment':
+            quantity = 0.0
+            for move in self.moves:
+                if move.state == 'done':
+                    quantity += Uom.compute_qty(move.uom, move.quantity,
+                        self.unit)
+            if self.quantity < 0:
+                quantity *= -1
+            return quantity
+
+    def _get_invoiced_quantity(self):
+        'Return the quantity already invoiced'
+        pool = Pool()
+        Uom = pool.get('product.uom')
+
+        quantity = 0
+        skips = {l for i in self.purchase.invoices_recreated for l in i.lines}
+        for invoice_line in self.invoice_lines:
+            if invoice_line.type != 'line':
+                continue
+            if invoice_line not in skips:
+                quantity += Uom.compute_qty(invoice_line.unit,
+                    invoice_line.quantity, self.unit)
+        return quantity
+
+    def _get_invoice_line_moves(self):
+        'Return the stock moves that should be invoiced'
+        moves = []
+        if self.purchase.invoice_method == 'order':
+            moves.extend(self.moves)
+        elif self.purchase.invoice_method == 'shipment':
+            for move in self.moves:
+                if move.state == 'done':
+                    if move.invoiced_quantity < move.quantity:
+                        moves.append(move)
+        return moves
+
+    def get_move(self, move_type):
         '''
         Return move values for purchase line
         '''
         pool = Pool()
-        Uom = pool.get('product.uom')
         Move = pool.get('stock.move')
 
         if self.type != 'line':
@@ -1312,14 +1312,14 @@ class PurchaseLine(ModelSQL, ModelView):
             return
         if self.product.type == 'service':
             return
-        skip = set(self.moves_recreated)
-        quantity = abs(self.quantity)
-        for move in self.moves:
-            if move not in skip:
-                quantity -= Uom.compute_qty(move.uom, move.quantity,
-                    self.unit)
 
-        quantity = Uom.round(quantity, self.unit.rounding)
+        if (self.quantity >= 0) != (move_type == 'in'):
+            return
+
+        quantity = (self._get_move_quantity(move_type)
+            - self._get_shipped_quantity(move_type))
+
+        quantity = self.unit.round(quantity)
         if quantity <= 0:
             return
 
@@ -1339,20 +1339,47 @@ class PurchaseLine(ModelSQL, ModelView):
         move.unit_price = self.unit_price
         move.currency = self.purchase.currency
         move.planned_date = self.delivery_date
-        move.invoice_lines = [l for l in self.invoice_lines
-            if not l.stock_moves]
+        move.invoice_lines = self._get_move_invoice_lines(move_type)
         move.origin = self
         return move
 
-    def create_move(self):
-        '''
-        Create move line
-        '''
-        move = self.get_move()
-        if not move:
-            return
-        move.save()
-        return move
+    def _get_move_quantity(self, move_type):
+        'Return the quantity that should be shipped'
+        return abs(self.quantity)
+
+    def _get_shipped_quantity(self, move_type):
+        'Return the quantity already shipped'
+        pool = Pool()
+        Uom = pool.get('product.uom')
+
+        quantity = 0
+        skip = set(self.moves_recreated)
+        for move in self.moves:
+            if move not in skip:
+                quantity += Uom.compute_qty(move.uom, move.quantity,
+                    self.unit)
+        return quantity
+
+    def _get_move_invoice_lines(self, move_type):
+        'Return the invoice lines that should be shipped'
+        return [l for l in self.invoice_lines if not l.stock_moves]
+
+    @classmethod
+    def view_attributes(cls):
+        return [
+            ('/form//field[@name="note"]|/form//field[@name="description"]',
+                'spell', Eval('_parent_purchase', {}).get('party_lang'))]
+
+    @classmethod
+    def copy(cls, lines, default=None):
+        if default is None:
+            default = {}
+        default = default.copy()
+        default['moves'] = None
+        default['moves_ignored'] = None
+        default['moves_recreated'] = None
+        default['invoice_lines'] = None
+        return super(PurchaseLine, cls).copy(lines, default=default)
 
 
 class PurchaseLineTax(ModelSQL):
@@ -1401,7 +1428,7 @@ class OpenSupplier(Wizard):
         ModelData = pool.get('ir.model.data')
         Wizard = pool.get('ir.action.wizard')
         Purchase = pool.get('purchase.purchase')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         purchase = Purchase.__table__()
 
         cursor.execute(*purchase.select(purchase.party,
@@ -1427,7 +1454,7 @@ class HandleShipmentExceptionAsk(ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         model = Table('ir_model')
         # Migration from 1.2: packing renamed into shipment
         cursor.execute(*model.update(
